#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Simple detection program for SmartDOC challenge 1: 
    ``Document object detection and segmentation in preview frames''

A quick way to start is to check and change the implementation for the 3 TODOs
below.

To invoke this program, use:
$ python process_sample.py /path/to/backgroundXX/documentNNN.avi -o /path/to/xml/output

To see invocation options, use:
$ python process_sample.py -h

You will need a working installation of:
- Python (>2.7)
- Numpy
- OpenCV (>2.4.8) with XVID support

"""

# ==============================================================================
# Imports
import logging
import argparse
import os
import os.path
import sys
import datetime
import cv,cv2
import numpy as np

# ==============================================================================
logger = logging.getLogger(__name__)

# ==============================================================================
# Constants
PROG_VERSION = "0.1"
PROG_NAME = "Simple detection program"

EXITCODE_OK = 0
EXITCODE_KBDBREAK = 10
EXITCODE_UNKERR = 254

# ==============================================================================
class FrameIterator(object):
    """
    Wrapper for OpenCV video file reader as a Python iterator.
    """
    def __init__(self, videofile):
        if not os.path.exists(videofile):
            err = "'%s' does not exist." % videofile
            logger.error(err)
            raise IOError(err)
        self._videocap = cv2.VideoCapture(videofile)
        self._frame_count = int(self._videocap.get(cv2.cv.CV_CAP_PROP_FRAME_COUNT))
        logger.debug("Input video informations:")
        logger.debug("\tframe_count = %d" % self._frame_count)
        self._cfid = 0
        self._prevRes = True

    def __iter__(self):
        return self

    def next(self):
        if self._prevRes and self._cfid < self._frame_count:
            self._prevRes, frame = self._videocap.read()
            self._cfid += 1
            if self._prevRes:
                return (self._cfid, frame)
        # else
        raise StopIteration

    def release(self):
        self._videocap.release()
        self._videocap = None

# ==============================================================================
class OutputDriver(object):
    """
    Very simple driver for XML result output.
    """
    def __init__(self):
        self.source_sample_file = ""
        self.software_used = ""
        self.software_version = ""
        self.segmentation_results = []

    def exportToFile(self, xmlfile):
        with open(xmlfile, "w") as out:
            out.write("<?xml version='1.0' encoding='utf-8'?>\n")
            out.write("<seg_result version=\"0.2\" generated=\"%s\">\n" % datetime.datetime.now().isoformat())
            out.write("  <software_used name=\"%s\" version=\"%s\"/>\n" % (self.software_used, self.software_version))
            out.write("  <source_sample_file>%s</source_sample_file>\n" % self.source_sample_file)
            out.write("  <segmentation_results>\n")
            for (fidx, rejected, tl, bl, br, tr) in self.segmentation_results:
                if rejected:
                    out.write("    <frame index=\"%d\" rejected=\"true\"/>\n" % fidx)
                else:
                    out.write("    <frame index=\"%d\" rejected=\"false\">\n" % fidx)
                    out.write("       <point name=\"bl\" x=\"%f\" y=\"%f\"/>\n" % (bl[0], bl[1]))
                    out.write("       <point name=\"tl\" x=\"%f\" y=\"%f\"/>\n" % (tl[0], tl[1]))
                    out.write("       <point name=\"tr\" x=\"%f\" y=\"%f\"/>\n" % (tr[0], tr[1]))
                    out.write("       <point name=\"br\" x=\"%f\" y=\"%f\"/>\n" % (br[0], br[1]))
                    out.write("    </frame>\n")
            out.write("  </segmentation_results>\n")
            out.write("</seg_result>\n")


# ==============================================================================
class Tracker(object):
    """
    A Tracker can process frames and find objects inside them.
    """

    def processFrame(self, fid, img):
        """
        Will be called once with each frame to process.
        """

        #
        # TODO reimplement this function
        #

        gray = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)
        blur = cv2.medianBlur(gray,75)
        img2 = blur

        (height, width) = img2.shape
        c = (width/2, height/2)

        THR = 180
        def go_dir(xdir, xlim, ydir, ylim):
            cx, cy = c
            c_ok = False
            while (cy + ydir != ylim and cx + xdir != xlim and not c_ok):
                if img2[cy + ydir, cx + xdir] > THR:
                    cy = cy + ydir ; cx = cx + xdir
                elif img2[cy + ydir, cx] > THR:
                    cy = cy + ydir
                elif img2[cy, cx + xdir] > THR:
                    cx = cx + xdir
                else:
                    c_ok = True
            return cx, cy

        tl = go_dir(-1,        0, -1,        0)
        bl = go_dir(-1,        0,  1, height-1)
        br = go_dir( 1,  width-1,  1, height-1)
        tr = go_dir( 1,  width-1, -1,        0)
        rejected = tl==c and bl==c and br==c and tr==c
        return (rejected, tl, bl, br, tr)


# ==============================================================================
def main(argv):
    # Option parsing
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description='Detects and segments document objects in frames.', 
        version=PROG_VERSION)

    parser.add_argument('-d', '--debug', 
        action="store_true", 
        help="Activate debug output.")

    parser.add_argument('-g', '--gui', 
        action="store_true", 
        help="Activate GUI to visualize tracker output.")

    parser.add_argument('-o', '--output-file', 
        help="Optionnal path to output file (XML description of segmentation).")

    parser.add_argument('input_sample', 
        help='Input sample.')

    args = parser.parse_args()

    # -----------------------------------------------------------------------------
    # Logger init
    format="%(name)-12s %(levelname)-7s: %(message)s" #%(module)-10s
    formatter = logging.Formatter(format)    
    ch = logging.StreamHandler()  
    ch.setFormatter(formatter)  
    logger.addHandler(ch)
    level = logging.INFO
    if args.debug:
        level = logging.DEBUG
    logger.setLevel(level)
    logger.debug("Arguments:")
    for (k, v) in args.__dict__.items():
        logger.debug("    %-20s = %s" % (k, v))

    # --------------------------------------------------------------------------
    # Prepare process
    logger.debug("Starting up")
    frames = FrameIterator(args.input_sample)

    # --------------------------------------------------------------------------
    # Initialize tracker
    logger.debug("Creating tracker...")
    tracker = Tracker()

    # --------------------------------------------------------------------------
    # Output preparation
    output = OutputDriver()
    output.source_sample_file = args.input_sample
    output.software_used = "My simple tracker" # TODO change name
    output.software_version = "0.1" # TODO change version

    # Let's go
    # --------------------------------------------------------------------------
    logger.debug("--- Process started. ---")
    # --------------------------------------------------------------------------
    win_name = 'Tracker output'
    if args.gui:
        cv2.namedWindow(win_name, cv2.WINDOW_NORMAL)

    for (fidx, mat) in frames:
        (frame_height, frame_width, _ch) = mat.shape
        (rejected, tl, bl, br, tr) = tracker.processFrame(fidx, mat)

        if not rejected:
            logger.info("frame %04d: A tl:(%-4.2f,%-4.2f) bl:(%-4.2f,%-4.2f) br:(%-4.2f,%-4.2f) tr:(%-4.2f,%-4.2f)" 
                        %(fidx, tl[0], tl[1], bl[0], bl[1], br[0], br[1], tr[0], tr[1]))
        else:
            logger.info("frame %04d: R" % fidx)

        # Debug viz
        if args.gui:
            if not rejected:
                    dgbq = np.int32([[tl[0], tl[1]],
                                     [bl[0], bl[1]],
                                     [br[0], br[1]],
                                     [tr[0], tr[1]]])
                    cv2.polylines(mat, [dgbq], True, (0, 255, 0), 2)

                    for txt, pt in zip(("tl", "bl", "br", "tr"), (tl, bl, br, tr)):
                        cv2.putText(mat, txt.upper(), (int(pt[0]), int(pt[1])), cv2.FONT_HERSHEY_PLAIN, 2, (64, 255, 64), 2)
            else:
                cv2.circle(mat, (frame_width/2, frame_height/2), 20, (0, 0, 255), 10)
                
            cv2.imshow(win_name, mat)
            key = cv2.waitKey(1) # required otherwise display thread show nothing
            if key & 0xFF == ord('q'):
                logger.warning("Exit requested, breaking up…")
                return EXITCODE_KBDBREAK

        output.segmentation_results.append((fidx, rejected, tl, bl, br, tr))

    # Output file
    if args.output_file is not None:
        output.exportToFile(args.output_file)
        logger.debug("SegResult file generated: %s" % args.output_file)

    # --------------------------------------------------------------------------
    logger.debug("--- Process complete. ---")
    # --------------------------------------------------------------------------
    if args.gui:
        logger.info("Press any key to exit.")
        cv2.waitKey()
        cv2.destroyWindow(win_name)


# ==============================================================================
# ==============================================================================
if __name__ == "__main__":
    ret = main(sys.argv)
    if ret is not None:
        sys.exit(ret)
